
#include "definition.h"

APP_TIMER_DEF(m_blink_usb);

void blink_handler(void * p_context);
void ble_midisend(uint8_t *data, byte datalen);

static TaskHandle_t m_usbd_thread;        /**< USB stack thread. */


// USB CODE START
static bool m_usb_connected = false;

extern bool isBLEmode;

/* ---------------------------- USB MIDI DEFINITION --------------------*/

// USB DEFINES START
/**
 * @brief USB Midi buffer size
 */
#define TX_BUFFER_SIZE 2048
#define SYSEX_BUF_SIZE 74

uint8_t m_sysex_buf[SYSEX_BUF_SIZE];
/**
 * @brief Enable power USB detection
 *
 * Configure if example supports USB port connection
 */
#ifndef USBD_POWER_DETECTION
#define USBD_POWER_DETECTION true
#endif

/**
 * @brief Midi class user event handler
 */

static void midi_user_ev_handler(app_usbd_class_inst_t const * p_inst,
                                    app_usbd_midi_user_event_t   event);

static void midi_user_rx_handler(app_usbd_class_inst_t const * p_inst,
                                    enum app_usbd_midi_rx_event_e event,
                                    uint8_t cable,
                                    app_usbd_midi_msg_t *rx);

/*
 * @brief   Midi class complete interface descriptor
 */
APP_USBD_MIDI_DESCRIPTOR(m_midi_desc,
                         APP_USBD_AUDIO_MIDI_CS_MIDI_STREAMING_INTERFACE_DSC,
                         APP_USBD_AUDIO_MIDI_EMBEDDED_IN_JACK_DSC,
                         APP_USBD_AUDIO_MIDI_EXTERNAL_IN_JACK_DSC,
                         APP_USBD_AUDIO_MIDI_EMBEDDED_OUT_JACK_DSC,
                         APP_USBD_AUDIO_MIDI_EXTERNAL_OUT_JACK_DSC,
                         APP_USBD_AUDIO_MIDI_STANDARD_BULK_OUT_ENDPOINT_DSC,
                         APP_USBD_AUDIO_MIDI_BULK_OUT_ENDPOINT_DSC,
                         APP_USBD_AUDIO_MIDI_STANDARD_BULK_IN_ENDPOINT_DSC,
                         APP_USBD_AUDIO_MIDI_BULK_IN_ENDPOINT_DSC);



/**
 * @brief Interfaces list passed to @ref APP_USBD_MIDI_GLOBAL_DEF
 */
#define MIDI_INTERFACES_CONFIG() APP_USBD_MIDI_CONFIG_IN_OUT(0, 1)

/*lint -save -e26 -e64 -e123 -e505 -e651*/

/**
 * @brief Midi class instance
 */
APP_USBD_MIDI_GLOBAL_DEF(m_app_midi,
                          MIDI_INTERFACES_CONFIG(),
                          midi_user_ev_handler,
                          midi_user_rx_handler,
                          &m_midi_desc,
                          TX_BUFFER_SIZE
);
// USB DEFINES END

/*--------------------------------------- DEFINITION OF USB MIDI HANDLER ---------------------*/

/**
 * @brief Function for midi rx handler
 *
 * @details midi isr event management
 */
static void midi_user_rx_handler(app_usbd_class_inst_t const * p_inst,
                                    enum app_usbd_midi_rx_event_e event,
                                    uint8_t cable,
                                    app_usbd_midi_msg_t *rx)
    {
    switch (event)
    {
    case APP_USBD_MIDI_SYSEX_BUF_REQ:
        {
            if (rx->p_data == NULL) 
            {
                /* requesting buffer for new sysex message */

                size_t len = SYSEX_BUF_SIZE;
                rx->p_data = m_sysex_buf;
                rx->len = len;
            }
            else
            {
                /* requesting new buffer because current buffer is full */

                // app_usbd_midi_sysex_write(&m_app_midi, 0, rx->p_data, rx->len);
                // NRF_LOG_HEXDUMP_INFO(rx->p_data, rx->len);

                size_t len = SYSEX_BUF_SIZE;
                rx->p_data = m_sysex_buf;
                rx->len = len;
            }
            return;
        }
    case APP_USBD_MIDI_SYSEX_RX_DONE:
        /* sysex message is completed */

        // app_usbd_midi_sysex_write(&m_app_midi, 0, rx->p_data, rx->len);
        // NRF_LOG_HEXDUMP_INFO(rx->p_data, rx->len);
        //ret = app_timer_stop(m_blink_usb);
        //APP_ERROR_CHECK(ret);
        break;
    case APP_USBD_MIDI_RX_DONE:
        /* received midi message */

        app_usbd_midi_write(&m_app_midi, 0, rx->p_data, rx->len);
        // NRF_LOG_HEXDUMP_INFO(rx->p_data, rx->len);

        NRF_LOG_INFO("MIDI Rx");
        return;
    default:
        return;
    }
}

/**
 * @brief User event handler @ref app_usbd_midi_user_ev_handler_t
 * */
static void midi_user_ev_handler(app_usbd_class_inst_t const * p_inst,
                                    app_usbd_midi_user_event_t event)
{
    switch (event)
    {
        case APP_USBD_MIDI_USER_EVT_CLASS_REQ:
            break;
        case APP_USBD_MIDI_USER_EVT_PORT_OPEN:
        {
            ret_code_t ret;
            NRF_LOG_INFO("MIDI port opened");
//            bsp_board_led_invert(BLE_LED);
            break;
        }
        case APP_USBD_MIDI_USER_EVT_PORT_CLOSE:
            NRF_LOG_INFO("MIDI Port Close Evt");
            break;
        case APP_USBD_MIDI_USER_EVT_TX_DONE:
            NRF_LOG_INFO("Tx done");
            break;

        default:
            break;
    }
}

/* ------------------------------------------------- END MIDI HANDLER ------------------------------------------------*/

/**
 * @brief Function for usb isr handler
 *
 * @details interrupt service routine by usb
 */
void usb_new_event_isr_handler(app_usbd_internal_evt_t const * const p_event, bool queued)
{
    BaseType_t xHigherPriorityTaskWoken = pdFALSE;
    UNUSED_PARAMETER(p_event);
    UNUSED_PARAMETER(queued);
    ASSERT(m_usbd_thread != NULL);
    /* Release the semaphore */
    vTaskNotifyGiveFromISR(m_usbd_thread, &xHigherPriorityTaskWoken);
    portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
}

/**
 * @brief Function for usbd handler
 *
 * @details track usbd events and perform task
 */
static void usbd_user_ev_handler(app_usbd_event_type_t event)
{
    switch (event)
    {
        case APP_USBD_EVT_DRV_SUSPEND:
        {
            ret_code_t err_code = app_timer_stop(m_blink_usb);
            APP_ERROR_CHECK(err_code);
        }    break;

        case APP_USBD_EVT_DRV_RESUME:
        {
            ret_code_t err_code = app_timer_stop(m_blink_usb);
            APP_ERROR_CHECK(err_code);
        }    break;

        case APP_USBD_EVT_STARTED:
            break;

        case APP_USBD_EVT_STOPPED:
            app_usbd_disable();
            break;

        case APP_USBD_EVT_POWER_DETECTED:
            NRF_LOG_INFO("USB power detected");

            if (!nrf_drv_usbd_is_enabled())
            {
                app_usbd_enable();
            }
            break;

        case APP_USBD_EVT_POWER_REMOVED:
        {
            NRF_LOG_INFO("USB power removed");
            m_usb_connected = false;
            app_usbd_stop();
        }
            break;

        case APP_USBD_EVT_POWER_READY:
        {
            NRF_LOG_INFO("USB ready");
            m_usb_connected = true;
          //  bsp_board_led_invert(BLE_LED);
            app_usbd_start();
        }
            break;

        default:
            break;
    }
}

/**
 * @brief Function for usbd thread
 *
 * @details initialize usbd handler and configure midi class.
 */
void usbd_thread(void * arg)
{
    ret_code_t ret;
    static const app_usbd_config_t usbd_config = {
        .ev_isr_handler = usb_new_event_isr_handler,
        .ev_state_proc  = usbd_user_ev_handler
    };
    UNUSED_PARAMETER(arg);

    ret = app_usbd_init(&usbd_config);
    APP_ERROR_CHECK(ret);
    app_usbd_class_inst_t const * class_inst_midi = app_usbd_midi_class_inst_get(&m_app_midi);
    ret = app_usbd_class_append(class_inst_midi);
    APP_ERROR_CHECK(ret);
    ret = app_usbd_power_events_enable();
    APP_ERROR_CHECK(ret);

    // Set the first event to make sure that USB queue is processed after it is started
    UNUSED_RETURN_VALUE(xTaskNotifyGive(xTaskGetCurrentTaskHandle()));
    // Enter main loop.
    for (;;)
    {
        /* Waiting for event */
        UNUSED_RETURN_VALUE(ulTaskNotifyTake(pdTRUE, USB_THREAD_MAX_BLOCK_TIME));
        while (app_usbd_event_queue_process())
        {
            /* Nothing to do */
        }
    }
}

/**
 * @brief Function for freeRTOS usb device task creation
 *
 * @details This function create freeRTOS task for usbd midi and timer.
 */
void create_usbd_thread (void)
{
    //usbd timer
    ret_code_t err_code = app_timer_init();
    err_code = app_timer_create(&m_blink_usb, APP_TIMER_MODE_REPEATED, blink_handler);
    APP_ERROR_CHECK(err_code);

    //create task for usbd 
    if (pdPASS != xTaskCreate(usbd_thread,
                              "USBD",
                              USBD_STACK_SIZE,
                              false,
                              USBD_PRIORITY,
                              &m_usbd_thread))
    {
        APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
    }
    NRF_LOG_INFO("USBD BLE UART example started.");
}

void SendMidiMessage (uint8_t *message, byte len)
{
      app_usbd_midi_write(&m_app_midi, 0, message, len);
     
#ifdef BLE_MIDI_SUPPORT
    if(isBLEmode)
    {
        ble_midisend(message, len);
    }
    else
#endif
    {
        app_usbd_midi_write(&m_app_midi, 0, message, len);

       // uart_send_midi_data(message, len);
    }
    
}