/**
 * Copyright (c) 2017 - 2021, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/** @file
 *
 * @defgroup usbd_ble_uart_freertos_example main.c
 * @{
 * @brief    USBD CDC ACM over BLE using FreeRTOS application main file.
 *
 * This file contains the source code for a sample application that uses the Nordic UART service
 * and USBD CDC ACM library and runs using FreeRTOS.
 * This application uses the @ref srvlib_conn_params module.
 */

#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "ble_hci.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_conn_params.h"
#include "nrf_sdh.h"
#include "nrf_sdh_soc.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_freertos.h"
#include "nrf_ble_gatt.h"
#include "app_timer.h"
#include "ble_nus.h"
#include "app_uart.h"
#include "app_util_platform.h"
#include "bsp_btn_ble.h"

#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#include "nrf_drv_usbd.h"
#include "nrf_drv_clock.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"
#include "nrf_drv_power.h"

#include "app_error.h"
#include "app_util.h"
#include "app_usbd_core.h"
#include "app_usbd.h"
#include "app_usbd_string_desc.h"
#include "app_usbd_cdc_acm.h"
#include "app_usbd_serial_num.h"

/* FreeRTOS related */
#include "nrf.h"                        // Device header
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

#include "global_user.h"
#include "looptimus_mini_plus_app.h"
#include "midi_ble.h"

#define APP_BLE_OBSERVER_PRIO               3                                          /**< Application's BLE observer priority. You shouldn't need to modify this value. */
#define APP_BLE_CONN_CFG_TAG                1                                          /**< A tag identifying the SoftDevice BLE configuration. */
      
extern bool isBLEmode;  
#ifdef  ONLY_HID_KEYBOARD_TESTING 
int xBLE_mode = 1;
#endif
int xBLE_mode = 0;
extern boolean gsInitDone;
int ble_mode_type = 0; // default BLE mode is MIDI

/*
  BLE MODE DETECT BY TRIGGERING A GPIO PIn
  0 = BLE midi & 1 = BLE keyboard 
*/
#define  DETECT_BLE_LOOPTIMUS_MODE() ((Detect_BLE_mode() == 1) ? (xBLE_mode = 1) : (xBLE_mode = 0))

BLE_MODE_SELECT ble_mode;

/**
 * The size of the stack for the Logger task (in 32-bit words).
 * Logger uses sprintf internally so it is a rather stack hungry process.
 */
#define LOGGER_STACK_SIZE 512
/**
 * The priority of the Logger task.
 */
#define LOGGER_PRIORITY 1
#define USB_THREAD_MAX_BLOCK_TIME portMAX_DELAY

/* DEFINITION OF USER DEFINED FUNCTION */
void user_timer_init(void);
void looptimus_mini_plus_app_thread(void * arg);
void mem_operation_thread (void *arg);
void init_looptimus(void);
void ble_init (void);
void advertising_start(void* p_context);
void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context);
void ble_evt_hid_key_handler(ble_evt_t const * p_ble_evt, void * p_context);
bool register_hid_keyboard_bond(void);
void ble_hid_init(void);
uint32_t buffer_dequeue(bool tx_flag);

void BLE_GAP_KEYBOARD_CONNECTED (ble_evt_t const * p_ble_evt);
void BLE_GAP_KEYBOARD_DISCONNECTED(void);
void NRF_REGISTER_MIDI_HANDLER(void);
#define LED_BLE_NUS_CONN (BSP_BOARD_LED_0)
#define LED_BLE_NUS_RX   (BSP_BOARD_LED_1)

APP_TIMER_DEF(m_blink_ble);

TaskHandle_t m_looptimus_app_thread;   
TaskHandle_t m_mem_operation_thread;   

#if NRF_LOG_ENABLED
static TaskHandle_t m_logger_thread;      /**< Logger thread. */
#endif

extern uint16_t   m_conn_handle;
/**
 * @brief App timer handler for blinking the LEDs.
 *
 * @param p_context LED to blink.
 */
void blink_handler(void * p_context)
{
   // bsp_board_led_invert((uint32_t) p_context);
}

#define ENDLINE_STRING "\r\n"

/** @brief Function for initializing the timer module. */
static void timers_init(void)
{
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);
    err_code = app_timer_create(&m_blink_ble, APP_TIMER_MODE_REPEATED, blink_handler);
    APP_ERROR_CHECK(err_code);
}

/**
 * @brief Function for handling errors from the Connection Parameters module.
 *
 * @param[in] nrf_error  Error code containing information about what went wrong.
 */
static void conn_params_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}

/**
 * @brief Function for handling BLE events (BLE MIDI and BLE Keyboard).
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
 /*
static void ble_evt_common_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
   uint32_t err_code;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            NRF_LOG_INFO("Connected");

            //BLE MIDI
            if((xBLE_mode == MIDI_BLE_MODE) && (xBLE_mode != KEYBOARD_BLE_MODE))
            {
                err_code = app_timer_stop(m_blink_ble);
                APP_ERROR_CHECK(err_code);
                bsp_board_led_on(LED_BLE_NUS_CONN);
                m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
                isBLEmode = true;                 //BLE Mode ON           
            }
            
            // BLE Keyboard
            else if ((xBLE_mode != MIDI_BLE_MODE) && (xBLE_mode == KEYBOARD_BLE_MODE))
            {    
                  BLE_GAP_KEYBOARD_CONNECTED(p_ble_evt);              
            }
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            NRF_LOG_INFO("BLE NUS disconnected");

            //BLE MIDI
            if((xBLE_mode == MIDI_BLE_MODE) && (xBLE_mode != KEYBOARD_BLE_MODE))
            {
                // LED indication will be changed when advertising starts.
                m_conn_handle = BLE_CONN_HANDLE_INVALID;
                bsp_board_led_off(LED_BLE_NUS_CONN);
                isBLEmode = false;                //BLE Mode OFF
            }

            // BLE Keyboard
            else if ((xBLE_mode != MIDI_BLE_MODE) && (xBLE_mode == KEYBOARD_BLE_MODE))
            {
                
                BLE_GAP_KEYBOARD_DISCONNECTED();
            }
            break;

        case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
        {
            NRF_LOG_DEBUG("PHY update request.");

            ble_gap_phys_t const phys =
            {
                    .rx_phys = BLE_GAP_PHY_AUTO,
                    .tx_phys = BLE_GAP_PHY_AUTO,
            };
            err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
            APP_ERROR_CHECK(err_code);
        } break;

        case BLE_GATTS_EVT_HVN_TX_COMPLETE:
        {
            // Send next key event
            (void) buffer_dequeue(true);
            break;
        }
        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:

            //BLE MIDI
            if((xBLE_mode == MIDI_BLE_MODE) && (xBLE_mode != KEYBOARD_BLE_MODE))
            {
                // Pairing not supported.
                err_code = sd_ble_gap_sec_params_reply(m_conn_handle, BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP, NULL, NULL);
                APP_ERROR_CHECK(err_code);
            }

            break;

        case BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST:
        #if 0
        {
            ble_gap_data_length_params_t dl_params;

            // Clearing the struct will effectively set members to @ref BLE_GAP_DATA_LENGTH_AUTO.
            memset(&dl_params, 0, sizeof(ble_gap_data_length_params_t));
            err_code = sd_ble_gap_data_length_update(p_ble_evt->evt.gap_evt.conn_handle, &dl_params, NULL);
            APP_ERROR_CHECK(err_code);
        } 
        #endif
        break;

        case BLE_GATTS_EVT_SYS_ATTR_MISSING:
            // No system attributes have been stored.
            err_code = sd_ble_gatts_sys_attr_set(m_conn_handle, NULL, 0, 0);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GATTC_EVT_TIMEOUT:
            // Disconnect on GATT Client timeout event.
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GATTS_EVT_TIMEOUT:
            // Disconnect on GATT Server timeout event.
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_EVT_USER_MEM_REQUEST:
            err_code = sd_ble_user_mem_reply(p_ble_evt->evt.gattc_evt.conn_handle, NULL);
            APP_ERROR_CHECK(err_code);
            break;

        case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
        #if 0
        {
            ble_gatts_evt_rw_authorize_request_t  req;
            ble_gatts_rw_authorize_reply_params_t auth_reply;

            req = p_ble_evt->evt.gatts_evt.params.authorize_request;

            if (req.type != BLE_GATTS_AUTHORIZE_TYPE_INVALID)
            {
                if ((req.request.write.op == BLE_GATTS_OP_PREP_WRITE_REQ)     ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_NOW) ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL))
                {
                    if (req.type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_WRITE;
                    }
                    else
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
                    }
                    auth_reply.params.write.gatt_status = APP_FEATURE_NOT_SUPPORTED;
                    err_code = sd_ble_gatts_rw_authorize_reply(p_ble_evt->evt.gatts_evt.conn_handle,
                                                               &auth_reply);
                    APP_ERROR_CHECK(err_code);
                }
            }
        } break; // BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST
        #endif
        

        default:
            // No implementation needed.
            break;
    }
}
*/
/**
 * @brief Function for handling events from the BSP module.
 *
 * @param[in]   event   Event generated by button press.
 */

void bsp_event_handler(bsp_event_t event)
{
    uint32_t err_code;
    switch (event)
    {
        case BSP_EVENT_SLEEP:
            sleep_mode_enter();
            break;

        case BSP_EVENT_DISCONNECT:
            err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            if (err_code != NRF_ERROR_INVALID_STATE)
            {
                APP_ERROR_CHECK(err_code);
            }
            break;

        case BSP_EVENT_WHITELIST_OFF:
            if (m_conn_handle == BLE_CONN_HANDLE_INVALID)
            {
                err_code = ble_advertising_restart_without_whitelist(&m_advertising);
                if (err_code != NRF_ERROR_INVALID_STATE)
                {
                    APP_ERROR_CHECK(err_code);
                }
            }
            break;

        default:
            break;
    }
}

/** @brief Function for initializing buttons and LEDs. */
static void BLE_MiDi_Handler_register(void)
{
    uint32_t err_code = bsp_init(BSP_INIT_LEDS, bsp_event_handler);
    APP_ERROR_CHECK(err_code);
}


/** @brief Function for initializing the nrf_log module. */
static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}

// USB CODE END


#if NRF_LOG_ENABLED
/**@brief Thread for handling the logger.
 *
 * @details This thread is responsible for processing log entries if logs are deferred.
 *          Thread flushes all log entries and suspends. It is resumed by idle task hook.
 *
 * @param[in]   arg   Pointer used for passing some arbitrary information (context) from the
 *                    osThreadCreate() call to the thread.
 */
static void logger_thread(void * arg)
{
    UNUSED_PARAMETER(arg);

    while (1)
    {
        NRF_LOG_FLUSH();

        vTaskSuspend(NULL); // Suspend myself
    }
}
#endif //NRF_LOG_ENABLED

/**@brief A function which is hooked to idle task.
 * @note Idle hook must be enabled in FreeRTOS configuration (configUSE_IDLE_HOOK).
 */
void vApplicationIdleHook( void )
{
#if NRF_LOG_ENABLED
     vTaskResume(m_logger_thread);
#endif
}
/*
void BLE_BSP_Handler_register(bool n_mode)
{
  if (n_mode == true)
  { 
    n_mode = false;
    buttons_leds_init();
  }
  else if (n_mode == false)
  {
    n_mode = true;
    register_hid_keyboard_bond();
  }
  else
  {
   
  }
}
*/
void MIDI_BLE_Init(void)
{
    // Initialize BLE MIDI protocol
    ble_init(); 
    // Start BKE MIDI Advertising
    nrf_sdh_freertos_init(advertising_start, NULL); 

}

/**
 * @brief Function for the SoftDevice initialization.
 *
 * @details This function initializes the SoftDevice and the BLE event interrupt.
 */
static void BLE_Stack_Init(void)
{   
    ret_code_t err_code;
    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(err_code);

    // Enable BLE stack.
    err_code = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

    NRF_REGISTER_MIDI_HANDLER();
}

/**
 * @brief Function for detecting BLE mode.
 *
 * @details This function detect BLE mode. It would be a BLE Keyboard or BLE MIDI
 */
int Detect_BLE_mode(void)
{
    nrf_gpio_cfg_input(BLE_MODE_DETECT, BUTTON_PULL);
    
    if(bsp_button_is_pressed(BLE_MODE_DETECT_BUTTON_IDX) == true)
    {
      // Midi
      ble_mode_type = MIDI_BLE_MODE;
    }
    else if (bsp_button_is_pressed(BLE_MODE_DETECT_BUTTON_IDX) == false)
    {
      //keyboard
      ble_mode_type = KEYBOARD_BLE_MODE;
    }

    return ble_mode_type;
}

void Looptimus_BLE_Init(void)
{
  // bsp_board_init(BSP_INIT_LEDS);
}

static void Init_BLE_Protocol(void)
{
    switch(xBLE_mode)
    {
        case MIDI_BLE_MODE:
        {
            BLE_MiDi_Handler_register();
            BLE_Stack_Init();
            ble_init();
            nrf_sdh_freertos_init(advertising_start,NULL);
            break;
        }

        case KEYBOARD_BLE_MODE:
        {
            BLE_Stack_Init();
            ble_hid_init();
            break;
        }
    }
}

/** @brief Application main function. */
int main(void)
{   
    DETECT_BLE_LOOPTIMUS_MODE();
    // bsp_board_led_invert(BLE_LED_NUM);
    
    ret_code_t ret;

    // Initialize.
    log_init();
    user_timer_init();
    timers_init();
    app_usbd_serial_num_generate();
    ret = nrf_drv_power_init(NULL);
    APP_ERROR_CHECK(ret);
    ret = nrf_drv_clock_init();
    APP_ERROR_CHECK(ret);

    // Initialize looptimus mini plus functionality 
    init_looptimus();
   
#if NRF_LOG_ENABLED
    // Start execution.
    if (pdPASS != xTaskCreate(logger_thread,
                              "LOGGER",
                              LOGGER_STACK_SIZE,
                              NULL,
                              LOGGER_PRIORITY,
                              &m_logger_thread))
    {
        APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
    }
#endif

    // create a freeRTOS thread for usbd
    create_usbd_thread();

    if (pdPASS != xTaskCreate(looptimus_mini_plus_app_thread,
                              "LOOPTIMUS_MINI_PLUS",
                              LOOPTIMUS_STACK_SIZE,
                              NULL,
                              LOOPTIMUS_PRIORITY,
                              &m_looptimus_app_thread))
    {
        APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
    }

    if (pdPASS != xTaskCreate(mem_operation_thread,
                              "MEMORY",
                              MEMORY_STACK_SIZE,
                              NULL,
                              MEMORY_OPERATION_TASK_PRIORITY,
                              &m_mem_operation_thread))
    {
        APP_ERROR_HANDLER(NRF_ERROR_NO_MEM);
    }

    gsInitDone = true;
    
    // Activate deep sleep mode.
    SCB->SCR |= SCB_SCR_SLEEPDEEP_Msk;
    Init_BLE_Protocol();
    bsp_board_led_invert(BLE_LED_NUM);
     bsp_board_led_invert(3);
    vTaskStartScheduler();

    for (;;)
    {
       APP_ERROR_HANDLER(NRF_ERROR_FORBIDDEN);
    }
}
/**
 * @}
*/
